#include "WPILib.h"
#include "ThunderVision.h"
#include "IOmap.h"
#include "Camera.h"

//const Camera::Target t_null = { 0, 0, 0, 0, 0};

Camera::Camera() {
	tilt = new Servo(PWM_CAMERA_TILT_SERVO);
	tv = new ThunderVision();
	target_found = false;
	for(int i=0;i<5;i++) {
//		targets[i] = t_null;
	}
}

Camera::~Camera() {
	// TODO what else needs cleaning up
	delete tv;
}

void Camera::GetNewImage() {
	// check for IsFreshImage()????
	int get = AxisCamera::GetInstance().GetImage(cam_image);
	if(get == 0) {
		// haven't even gotten an image? certainly can't go on that way!
		best_t_id = none;
		target_found = false;
		return;
	}
	image = (RGBValue*)imaqImageToArray(cam_image,IMAQ_NO_RECT,NULL,NULL);
	int targets = tv->ProcessImage(image);
	
	IDTargets(targets);
}

float Camera::HeightOfTarget() {
	return targets[0].h;
}
float Camera::WidthOfTarget() {
	return targets[0].h;
}
float Camera::TargetCenterHeight() {
	switch (best_t_id) {
	case top: return TOP_HEIGHT;
	case center_left:
	case center_right:
		return CENTER_HEIGHT;
	case bottom: return BOTTOM_HEIGHT;
	default: return 0;
	}
}
float Camera::TargetCenterYPos() {
	switch (best_t_id) {
	case top:
	case bottom: return CENTRAL_YPOS;
	case center_left:
		return LEFT_YPOS;
	case center_right:
		return RIGHT_YPOS;
	default: return 0;
	}
}

bool Camera::TargetFound() {
	return target_found;
}
bool Camera::TargetCentered() {
	// absoute value without abs()
	int tmp = TargetHorizDelta();
	return (tmp < MAX_DELTA_X) && (tmp > -MAX_DELTA_X);
}
bool Camera::TargetIdentified() {
	return target_found && best_t_id!=none;
}
int Camera::TargetHorizDelta() {
	return IMAGE_CENTER_H - targets[0].c_x;
}
int Camera::TargetTilt() {
	return targets[0].tilt;
}

void Camera::SetTilt(float angle) {
    // Servo::SetAngle is set up to never overshoot its bounds
	tilt->SetAngle(angle+CAM_ANGLE_CORR);
	camAngle = angle;
}
float Camera::GetTilt() {
	return camAngle;
}
bool Camera::TiltDone() {
	return ((tilt->GetAngle()-CAM_ANGLE_CORR) == camAngle);
}

/* private functions */
void Camera::IDTargets(int count) {
printf("Targets:%d\n",count);
	// better to have a static array with sufficient capacity for all
	// but the worst case??
	ThunderVision::TARGET_INFO *ti = new ThunderVision::TARGET_INFO[count];
	tv->GetTargetInfo(ti,count);

	// number of targets, index of best target
	int n_tgts=0, bt=0;
	// convert TV format to TgtID format, stripping out noise in the process
	for(int i=0;i<count;i++) {
		// initial best (distance from center)^2 is initialized way high on purpose
		float x,y,d2,d2old=1000000.0;
		if(ti[i].maxWidth>MIN_WIDTH || ti[i].maxHeight>MIN_HEIGHT) {
			// it's probably not noise then!
			
			// horizontal and vertical center points
			targets[n_tgts].c_x = ti[i].right-ti[i].left;
			targets[n_tgts].c_y = ti[i].top-ti[i].bottom;
			// target height and width
			// TODO maxWidth is certainly preferable to avgWidth, but
			//      what about avgHeight vs. maxHeight? 
			targets[n_tgts].h = ti[i].avgHeight;
			targets[n_tgts].w = ti[i].maxWidth;
			if(ti[i].edgeTop>TILT_TOL) // tilts up to right  
				targets[n_tgts].tilt=+1;
			else if(ti[i].edgeTop<-TILT_TOL) // tilts up to left
				targets[n_tgts].tilt=-1;
			else targets[n_tgts].tilt=0;
			
			// calculate distance from center to see how much closer this
			// target is to center of image than previous best
			// TODO use of ti[i].row ???!???
			x = IMAGE_CENTER_H-targets[n_tgts].c_x;
			y = IMAGE_CENTER_V-targets[n_tgts].c_y;
			d2 = x*x+y*y;
			if(d2<d2old) { d2old=d2; bt=n_tgts; }
			n_tgts++; // found one!
		}
	}
	// clear any remaining targets in the array TODO is this necessary?
//	for(int i=n_tgts;i<5;i++) targets[i]=t_null;
	// interpret the data
	if(n_tgts==0) {
		best_t_id = none;
		target_found = false; // we don't have any targets
	}
	else if(n_tgts==1) {
		best_t_id = none; // we found one but can't ID it without ref.pts.
		target_found = true;
	}
	else { 
		// TODO would TARGET_INFO.row help us here and in CompareTargets?????
		target_found = true; // we found a target; now try to identify it
		// move best target to #0 (for ease of use??)
		targets[4] = targets[0];
		targets[0] = targets[bt];
		targets[bt] = targets[4];
		if(n_tgts==2) {
			// simple; see CompTgts function below
			best_t_id = CompareTargets(0,1);
		} else if(n_tgts==3) {
			// different method; "our" target is to right/left/above/below the other
			// two, just need to find out which direction is the case
			int tmp=0; // 1:right of t[1], 2:to left, 4:above; 8:below
			float d = targets[0].c_x - targets[1].c_x;
			if(d>targets[0].w*MEAN_RATIO) //right of t[1]?
				tmp&=1;
			else if(d<-targets[0].w*MEAN_RATIO) // left?
				tmp&=2;
			d = targets[0].c_y - targets[1].c_y;
			if(d>targets[0].h*MEAN_RATIO) // above?
				tmp&=4;
			else if(d<-targets[0].h*MEAN_RATIO) // below?
				tmp&=8;
			//now compare with other target, depending on previous results
			// the first two conditions are mutually exclusive, as are the last 2!
			if(tmp&1 && (targets[0].c_x-targets[2].c_x)>(targets[0].w*MEAN_RATIO))
				best_t_id = center_right; // right of all the others
			else if(tmp&2 && (targets[0].c_x-targets[2].c_x)<(-targets[0].w*MEAN_RATIO))
				best_t_id = center_left; // left of all the others
			else if(tmp&4 && (targets[0].c_y-targets[2].c_y)>(targets[0].h*MEAN_RATIO))
				best_t_id = top; // above all the others
			else if(tmp&8 && (targets[0].c_y-targets[2].c_y)<(-targets[0].h*MEAN_RATIO))
				best_t_id = bottom; //below all the others
			else best_t_id = none; // inconclusive (and unlikely)
		} else { //n_tgts==4
			// look for an extreme combination to ID the target
			for(int i=1;i<4;i++) {
				best_t_id = CompareTargets(0,i);
				if(best_t_id!=none) break; //possibly save some time!
			}
		}
	} // else [count>1]; best_t_id and target_found will have been set in any case
}

Camera::TargetIdentity Camera::CompareTargets(int t1, int t2) {
	// only two tgt combos can be surely analyzed by this method: top+bottom or both centers
	float d = targets[t1].c_x - targets[t2].c_x;
	if(d>targets[t1].w*EXTREME_RATIO) // far enough right of other target?
		return center_right;
	else if(d<-targets[t1].w*EXTREME_RATIO) // far enough left?
		return center_left;
	
	d = targets[t1].c_y - targets[t2].c_y;
	if(d>targets[t1].h*EXTREME_RATIO) // far enough above?
		return top;
	else if(d<-targets[t1].h*EXTREME_RATIO) // far enough below?
		return bottom;
	else return none; // we can't tell
}
