#include "Controls.h"
#include "DSmap.h"

//if defined, allows display of rough target speed on
// dashboard for drivers (based on camera angle)
#define CAMANGLE_SPEED_FEEDBACK

#define SHOOT_SPEED_MOD 1.0	//Multiplier for shooter wheel speeds (0->1 * SHOOT_SPEED_MOD)
#define SHOOT_SPEED_DIFFERENCE 0.0 //Speed difference between the shooter wheels for backspin

#define TURRET_MANUAL_TOLERANCE 0.2

/*Things to keep in mind
 * We will need to do some things in automode
 * These will need to be seperate from the normal loop
 * 
 * LOTS OF VALUES WILL NEED TO CHANGE WHEN WE CAN TEST ON THE BOT 
 */ 


Controls::Controls(SmartShooter* shooter, Storage* storage, ThunderDrive* drive, Light* light, Joystick* leftJoystick, Joystick* rightJoystick)
{
	joyLeft = leftJoystick;
	joyRight = rightJoystick;
	extendedIO = new Joystick(DS_JOY_PORT_EMULATED);
	camLightRelay = new Relay(RELAY_CAMERA_LIGHT, Relay::kForwardOnly);
	driverStation = DriverStation::GetInstance();
	driverStationLCD = DriverStationLCD::GetInstance();
	dashboard = &driverStation->GetLowPriorityDashboardPacker();
	thunderShooter = shooter;
	thunderStorage = storage;
	thunderDrive = drive;
	thunderLight = light;
	lastLightButtonState = false;
}

Controls::~Controls()
{
	delete extendedIO;
	delete joyRight;
	delete joyLeft;
	delete camLightRelay;
}

void Controls::runControls()
{
	processBrokenSwitches();
	processJoysticks();	
	
}

float Controls::getPower(Joystick * stick)
{
	float power = stick->GetY();
	if(power > -.05 && power < .05)	//Deadzone, prevents motors from running
		power = 0;  				//when not touching the joysticks
	
	power = power * power * power;  //Cube the power to ramp input
	
	if(stick->GetRawButton(DS_JOYBTN_SLOW))
		power *= .3; //Cut power to 3/10 for slower driving
	else if(stick->GetRawButton(DS_JOYBTN_TURBO)){}
		//Do nothing, keep full power
	else
		power *= .7; //Cut power to 7/10 for normal operation
	
	return power;
}

void Controls::processJoysticks()
{
	thunderDrive->setMotorSpeed(getPower(joyLeft), getPower(joyRight)); //Set Motor Speeds
	
	if(!(extendedIO->GetRawButton(DS_BROKEN_TURRET_ANGLE)))	//Broken shooter angle?
		if(extendedIO->GetRawAxis(DS_TURRET_ANGLE) < 0)	//Turning left
			thunderShooter->setTurretAngle(extendedIO->GetRawAxis(DS_TURRET_ANGLE) * -TURRET_MAX_DEG); //THIS WILL NEED TO CHANGE
		else	//Turning right
			thunderShooter->setTurretAngle(extendedIO->GetRawAxis(DS_TURRET_ANGLE) * TURRET_MIN_DEG);
	else //turret pid control is broken, use manual mode
	{
		float x = extendedIO->GetRawAxis(DS_TURRET_ANGLE);
		if(extendedIO->GetRawAxis(DS_TURRET_ANGLE) > TURRET_MANUAL_TOLERANCE) {
		 	thunderShooter->moveTurret(x/2);
		} else if(extendedIO->GetRawAxis(DS_TURRET_ANGLE) < -TURRET_MANUAL_TOLERANCE)
		 	thunderShooter->moveTurret(x/2);
		else
		 	thunderShooter->stopTurret();
	}

	if (extendedIO->GetRawButton(DS_BALL_COLLECT_DOWN)) //Down, Out
	{ 
		thunderStorage->beaterBarMove(-1);
		thunderStorage->manualConvMove(-1);
	}
	else if (extendedIO->GetRawButton(DS_BALL_COLLECT_UP)){ //Up, In/Auto
		if(extendedIO->GetRawButton(DS_BROKEN_CONVEYOR))
		{
			thunderStorage->beaterBarMove(1);
			thunderStorage->manualConvMove(1);
		}
		else
		{
			thunderStorage->beaterBarMove(0);
			thunderStorage->manualConvMove(0);
		}
	}
	else //Center, off
	{
		thunderStorage->beaterBarMove(-100);
		thunderStorage->manualConvMove(-100);	    
	}
	
	if ((extendedIO->GetRawButton(DS_BROKEN_BRIDGE)))
	{
		if (joyRight->GetRawButton(DS_JOYBTN_WINGS_DOWN))
			thunderStorage->wingsMove(-1);
		else if (joyRight->GetRawButton(DS_JOYBTN_WINGS_UP))
			thunderStorage->wingsMove(1);
		else thunderStorage->wingsMove(0);
	}
	else if(joyRight->GetRawButton(DS_JOYBTN_WINGS))
		thunderStorage->setWingsPosition(Storage::WINGS_DOWN);
	else if(joyLeft->GetRawButton(DS_JOYBTN_WINGS))
		thunderStorage->setWingsPosition(Storage::WINGS_HALF);
	else
		thunderStorage->setWingsPosition(Storage::WINGS_UP);
		
	// Get top/bottom ratio
	// Bottom is always full input
	// Top varied against bottom as 0 -> 1.0 fraction of top
	float topToBottom;
#ifdef VARIABLEBACKSPIN
	topToBottom = extendedIO->GetRawAxis(DS_DEBUG_DIAL1); // -1->1
	topToBottom += 1;  // 0 -> 2
	topToBottom /= 2;  // 0 -> 1
#else
	topToBottom = 1.0f;
#endif
	
	float speed = extendedIO->GetRawAxis(DS_SHOOTER_POWER);
	speed = (speed+1)/2;
	if(!extendedIO->GetRawButton(DS_BROKEN_SMART_SHOOTER)) {
		thunderShooter->tiltCamera(speed * 90);
	}
	else
	{
		thunderShooter->setNormalizedSpeed((speed - SHOOT_SPEED_DIFFERENCE) * SHOOT_SPEED_MOD * topToBottom, speed * SHOOT_SPEED_MOD);
		thunderShooter->tiltCamera((extendedIO->GetRawAxis(DS_CAMTILT_DIAL)+1)*45);
	}
	
	INT8 raw = driverStation->GetStickAxisRaw(DS_JOY_PORT_EMULATED, DS_DIGITAL2);
	bool shoot = (raw & DS_FIRE) >> 5;
	if(shoot)
		thunderShooter->shoot();
	else
		thunderShooter->shootStop();

	bool autoAim = (driverStation->GetStickAxisRaw(DS_JOY_PORT_EMULATED, DS_DIGITAL2) & DS_AUTO_AIM) >> 6;
#if 0
	if (!extendedIO->GetRawButton(DS_BROKEN_SMART_SHOOTER)) {
		thunderShooter->setTargeting(autoAim);
		thunderLight->setShooting(autoAim);
	} else
#endif
	{
		thunderShooter->setTargeting(false);
		thunderLight->setShooting(false);
		camLightRelay->Set(autoAim ? Relay::kOn : Relay::kOff);
	}
	if (autoAim && joyLeft->GetRawButton(DS_JOYBTN_WINGS_DOWN))
		thunderStorage->setBridgeStabilizer(1.0f);
	else if (autoAim && joyLeft->GetRawButton(DS_JOYBTN_WINGS_UP))
		thunderStorage->setBridgeStabilizer(-0.8f);
	else
		thunderStorage->setBridgeStabilizer(0);
	
	thunderLight->setCoopertition(extendedIO->GetRawButton(DS_REQUEST_COOPERTITION));
	thunderShooter->setUseBackspin(extendedIO->GetRawButton(DS_ENABLE_BACKSPIN));
}

void Controls::processBrokenSwitches()
{
	thunderStorage->setWingsBroken(extendedIO->GetRawButton(DS_BROKEN_BRIDGE));
	thunderStorage->setConvBroken(extendedIO->GetRawButton(DS_BROKEN_CONVEYOR));
	thunderStorage->setCountBroken(extendedIO->GetRawButton(DS_BROKEN_CONVEYOR));
	thunderShooter->setTurretBroken(extendedIO->GetRawButton(DS_BROKEN_TURRET_ANGLE));
	thunderShooter->setWheelsBroken(extendedIO->GetRawButton(DS_BROKEN_SHOOTER_SPEED));
	// making variable wheels the default, "unbroken" position;
	// continuous is the "broken option".
	thunderShooter->setWheelsVariable(!extendedIO->GetRawButton(DS_CONTINUOUS_SHOOTER));
	//Broken Smart Shooter? 
	thunderShooter->setTargeting(false);
}

void Controls::getAutoValues(int* startPos, int* steps, int* delay)
{
	INT8 rawAnalog1 = driverStation->GetStickAxisRaw(DS_JOY_PORT_EMULATED, DS_DIGITAL1);
	INT8 rawAnalog2 = driverStation->GetStickAxisRaw(DS_JOY_PORT_EMULATED, DS_DIGITAL2);
	
	*startPos = (rawAnalog1 & DS_AUTOMODE_START_POS) >> 1;
	*steps = (rawAnalog1 & DS_AUTOMODE_STEPS) >> 3;
	*delay = (rawAnalog2 & DS_AUTOMODE_DELAY) >> 1;	 
}

void Controls::processDebug()
{
//	static int throttle = 0;
//	if ((throttle++ % 200) == 0)
//		printf("RAW: %d\n", (int)(driverStation->GetStickAxisRaw(DS_JOY_PORT_EMULATED, DS_DIGITAL2) & 0xff));

	short buttonStates = driverStation->GetStickButtons(DS_JOY_PORT_LEFT) & 0xFFE0;
	
	switch(buttonStates){
	case 0x0020:	//Button 6 - Drivetrain debug
		driverStationLCD->Clear();
		driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line1, "Left Encoder:  %i", thunderDrive->getEncoder(false));
		driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line2, "Right Encoder: %i", thunderDrive->getEncoder(true));
		driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line3, "Gyro: %f", (double)thunderDrive->getGyro());
		driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line5, "Harpoon Locked Up: %i", thunderStorage->isBridgeStabilizerLocked());
		driverStationLCD->UpdateLCD();
		break;
	case 0x0040:	//Button 7 - Shooter
		driverStationLCD->Clear();
		driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line1, "T S: %f", (double)thunderShooter->getTopSpeed());
		driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line2, "B S: %f", (double)thunderShooter->getBottomSpeed());
		driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line3, "T E: %i R: %f", thunderShooter->getTopEnc(), thunderShooter->getTopRev());
		driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line4, "B E: %i R: %f", thunderShooter->getBottomEnc(), thunderShooter->getBottomRev());
		driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line5, "Cam Angle: %f", (double)thunderShooter->getCamTilt());
		
		driverStationLCD->UpdateLCD();
		break;
	case 0x0080:	//Button 8 - Storage
	    // four conveyor sensors, both states and voltages, wings state, wings normalized value, wings raw value, wings mode, beater bar direction
	    char * wingsState;
	    switch (thunderStorage->getWingsPosition())
	    {
	    case 0:
	        wingsState = "Mid";
	        break;
	    case 1:
	        wingsState = "Top";
	        break;
	    case -1:
	        wingsState = "Bottom";
	        break;
	    case 2:
	    	wingsState = "Midway";
	    	break;
	    default:
	    	wingsState = "Something's wrong";
	    	break;
	    }
	    
	    char * beaterBarDir;
	    switch (thunderStorage->getBeaterBarMode())
	    {
	    case 1:
	        beaterBarDir = "In-Manual";
	        break;
	    case 0:
	        beaterBarDir = "Auto-Movement";
	        break;
	    case -1:
	        beaterBarDir = "Out-Manual";
	        break;
	    default:
	        beaterBarDir = "Stopped";
	    }
	    
	    driverStationLCD->Clear();
	    driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line1, "Wings: Raw: %fV, Norm: %f, State: %s", thunderStorage->wingsRawValue(), thunderStorage->wingsPos(), wingsState);    
        driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line2, "Beater Dir: %s", beaterBarDir);
        driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line3, "Sensor 0: %d %fV", thunderStorage->getSensorState(0), thunderStorage->getSensorVoltage(0));
        driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line4, "Sensor 1: %d %fV", thunderStorage->getSensorState(1), thunderStorage->getSensorVoltage(1));
        driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line5, "Sensor 2: %d %fV", thunderStorage->getSensorState(2), thunderStorage->getSensorVoltage(2));
        driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line6, "Sensor 3: %d %fV", thunderStorage->getSensorState(3), thunderStorage->getSensorVoltage(3));
		driverStationLCD->UpdateLCD();
		break;
	case 0x0100:	//Button 9
        driverStationLCD->Clear();
        driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line1, "Turret Angle: %f", (double)thunderShooter->getTurretAngle());
        driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line2, "Goto Angle: %f", (double)thunderShooter->getTurretDesiredAngle());
        driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line3, "Err Pct: %f", thunderShooter->getTurretError());
        driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line4, "Ticks: %d %f", thunderShooter->getTurretEnc(), thunderShooter->getTurretEncNorm());
        driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line5, "Lim Fwd (CW ): %f", (double)thunderShooter->getLimitCW());
        driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line6, "Lim Rev (CCW): %f", (double)thunderShooter->getLimitCCW());
        driverStationLCD->UpdateLCD();
        //printf("Turret Encoder: %i  CW: %f  CCW: %f\n", thunderShooter->getTurretEnc(), (double)thunderShooter->getLimitCW(), (double)thunderShooter->getLimitCCW());
		break;
	case 0x0200:	//Button 10 USED FOR LIGHT PATTERN SELECTION
		break;
	case 0x0400:	//Button 11 USED FOR LIGHT PATTERN SELECTION
		break;
	default:		//No debug buttons
		driverStationLCD->Clear();
		driverStationLCD->UpdateLCD();
		break;
	}	
	if(driverStation->GetAlliance() == DriverStation::kRed)
		thunderLight->setAlliance(true);
	if(driverStation->GetAlliance() == DriverStation::kBlue)
		thunderLight->setAlliance(false);

    thunderShooter->setLoggingEnabled(joyLeft->GetRawButton(DS_JOYBTN_DEBUG_SHOOTER));
}

void Controls::processLightControls()
{
	if(joyLeft->GetRawButton(DS_JOYBTN_NEXT_PATTERN) && !lastLightButtonState)
	{
		thunderLight->nextPattern();
	}
	lastLightButtonState = joyLeft->GetRawButton(DS_JOYBTN_NEXT_PATTERN);
	if(joyLeft->GetRawButton(DS_JOYBTN_SHOW_POS))
		thunderLight->showBallPos();
}

void Controls::processDashboard()
{
	dashboard->AddCluster();
		dashboard->AddI8((INT8)thunderStorage->getBallCount());					//Expects 0 to 3
		dashboard->AddFloat(thunderShooter->getTopSpeed());						//Expects 0 to 1
		dashboard->AddFloat((extendedIO->GetRawAxis(DS_SHOOTER_POWER)+1)/2);	//Expects 0 to 1
		dashboard->AddFloat(thunderShooter->getBottomSpeed());					//Expects 0 to 1
		dashboard->AddFloat((extendedIO->GetRawAxis(DS_SHOOTER_POWER)+1)/2);	//Expects 0 to 1
		dashboard->AddFloat((thunderShooter->getTurretAngle()+200)/3.2f);		//Expects 0 to 100
		dashboard->AddFloat((extendedIO->GetRawAxis(DS_TURRET_ANGLE)+1)*50);	//Expects 0 to 100
#ifdef CAMANGLE_SPEED_FEEDBACK
        dashboard->AddFloat(thunderShooter->getRoughTargetSpeed());             //Expects 0 to 1
#endif
	dashboard->FinalizeCluster();
	dashboard->Finalize();
}

