#ifndef CONTROLS_H
#define CONTROLS_H

#include "DriverStation.h"
#include "WPILib.h"
#include "SmartShooter.h"
#include "Storage.h"
#include "ThunderDrive.h"
#include "Light.h"

class Controls {
public:
	Controls(SmartShooter* shooter, Storage* storage, ThunderDrive* drive, Light* thunderLight, Joystick* leftJoystick, Joystick* rightJoystick);
	~Controls();
	/* Name: runControls
	 * Arguments: none
	 * Actions: Gets input from controls and calls 
	 * 			other robot functions (like setting motor speeds)
	 * Returns: squat
	 */
	void runControls(); 
	
	/* Name: getAutoValues
	 * Arguments: Pointers to startPos, steps, and delay in seconds
	 * Actions: Passes automode selection from control box at start of match
	 * Returns: nothing
	 */
	void getAutoValues(int* startPos, int* steps, int* delay);
	
	/* Name: processBrokenSwitches
	 * Arguments: Politics and religion
	 * Actions: Sets broken flags for rest of robot
	 * Returns: Nothing
	 */
	void processBrokenSwitches();
	
	/* Name: processDebug
	 * Arguments: none
	 * Actions: Prints debug text to the LCD on the driver station
	 * 			when debug buttons are pressed
	 * Returns: Nothing
	 */
	void processDebug();
	
	//Packs and sends data to the dashboard
	void processDashboard();
	
	/* Name: processLightControls
	 * Arguments: None
	 * Actions: reads the joystick buttons to see if we should change light patterns
	 * Returns: Nothing
	 */
	void processLightControls();
	
private:
	Joystick* joyLeft;
	Joystick* joyRight;
	Joystick* extendedIO;
	
	DriverStation* driverStation;
	DriverStationLCD* driverStationLCD;
	Dashboard* dashboard;
	
	SmartShooter* thunderShooter;
	Storage* thunderStorage;
	ThunderDrive* thunderDrive;
	Light* thunderLight;
	Relay* camLightRelay;
	
	bool lastLightButtonState;
	
	/* Name: getPower 
	 * Arguments: pointer to a joystick object
	 * Actions: Reads joystick buttons and Y value to 
	 * 			get the proper motor power
	 * Returns: -1 to 1, scaled power value for motors
	 * 			(Pass directly to setMotorSpeed)
	 */
	float getPower(Joystick * stick);
	
	//Processes input from operator joysticks
	void processJoysticks();
};

#endif
