#include "Light.h"

#define NORMAL_PWMRATE 1000.0

Light::Light(Storage* storage, GenericHID* left, GenericHID* right):
lightRed(DIG_LIGHT_RED),
lightGreen(DIG_LIGHT_GREEN),
lightBlue(DIG_LIGHT_BLUE),
lightSelect1(DIG_LIGHT_SELECT_1),
lightSelect2(DIG_LIGHT_SELECT_2),
lightSelect3(DIG_LIGHT_SELECT_3),
timer()
{
	thunderStorage = storage;
	kinectLeft = left;
	kinectRight = right;
	
	signalRequest = false;
	requestNextPattern = false;
	lightShow = ballPos;
	
	lightRed.SetPWMRate(NORMAL_PWMRATE);
	lightRed.EnablePWM(0);
	lightGreen.EnablePWM(0);
	lightBlue.EnablePWM(0);	
	alliance = false;
	
	timer.Start();
	
	srandom((unsigned long)GetTime()); //Seed the random number functions!
}

Light::~Light()
{
	
}

void Light::showStartupWarning(bool enable)
{
    if (enable) {
        lightRed.SetPWMRate(4.0f);
        setColor(1.0f, 0, 0);
        setLight(true, true, true);
    } else {
        lightRed.SetPWMRate(NORMAL_PWMRATE);
        setLight(false, false, false);
    }
}

void Light::processLights()
{
	if(thunderStorage->getBallCount())
	{
		//Set which lights are lit
		setLight(thunderStorage->getSensorState(1), thunderStorage->getSensorState(2), thunderStorage->getSensorState(3));
	}
	else
	{
		//No balls
		//Flash lights in sequence
		double curTime = timer.Get();

		if(curTime > 1)
			timer.Reset();
		else if(curTime > .75)
			setLight(false, true, false);
		else if(curTime > .5)
			setLight(false, false, true);
		else if(curTime > .25)
			setLight(false, true, false);
		else //if(curTime > 0)
			setLight(true, false, false);
	}
		
		//Set the color of the lights
	if(shooting)  //Are we targeting?
	{
		setColor(0,0,1);  //Blue
	}
	else if(thunderStorage->getWingsPosition()!=1 && thunderStorage->getWingsPosition()!=-1)  //Are wings not up and not down
	{
		setColor(0,1,1);  //Light Blue
	}
	else if(thunderStorage->getWingsPosition()!=1)  //Are wings down?
	{
		setColor(1,0,0);  //Red
	}
	else if(signalRequest) //Are we signaling for coopertition?
	{
		setColor(1,1,1);  //White
	}
	else if (thunderStorage->getBallCount())
	{
		setColor(0.2, 0, 0.38); //Purple
	}
	else  //Just balls
	{
		setColor(1,1,0);  //Yellow
	}
}

// Assumes full saturation and value
static void hueToRGB(float *rp, float *gp, float *bp, float h) {
    float s = 1.0f;
    float v = 1.0f;

    int i = (int)(h * 6);
    float f = h * 6 - i;
    float p = v * (1 - s);
    float q = v * (1 - f * s);
    float t = v * (1 - (1 - f) * s);
    float r,g,b;

    switch (i % 6){
        case 0: r = v; g = t; b = p; break;
        case 1: r = q; g = v; b = p; break;
        case 2: r = p; g = v; b = t; break;
        case 3: r = p; g = q; b = v; break;
        case 4: r = t; g = p; b = v; break;
        case 5: r = v; g = p; b = q; break;
    }

    *rp = r;
    *gp = g;
    *bp = b;
}

void Light::processLightShow()
{
	double curTime = timer.Get();
	float leftArm = (kinectLeft->GetY()+1)/2;
	float rightArm = (kinectRight->GetY()+1)/2;
	
	switch(lightShow){		
	case ballPos:
		if(requestNextPattern) 
		{
			lightShow = rider1;
			requestNextPattern = false;
		}
		
		if(alliance)
			setColor(1,0,0);
		else
			setColor(0,0,1);
		
		if(thunderStorage->getBallCount())
			setLight(thunderStorage->getSensorState(1), thunderStorage->getSensorState(2), thunderStorage->getSensorState(3));
		else
		{
			if(curTime > 1)
				timer.Reset();
			else if(curTime > .75)
				setLight(false, true, false);
			else if(curTime > .5)
				setLight(false, false, true);
			else if(curTime > .25)
				setLight(false, true, false);
			else //if(curTime > 0)
				setLight(true, false, false);
		}
		break;
		
	case rider1:
		if(requestNextPattern) 
		{
			lightShow = rider2;
			requestNextPattern = false;
		}
		if(alliance)
			setColor(1,0,0);
		else
			setColor(0,0,1);
		if(curTime > 1)
			timer.Reset();
		else if(curTime > .75)
			setLight(false, true, false);
		else if(curTime > .5)
			setLight(false, false, true);
		else if(curTime > .25)
			setLight(false, true, false);
		else
			setLight(true, false, false);
		break;

	case rider2:
		if(requestNextPattern) 
		{
			lightShow = kinect1;
			requestNextPattern = false;
		}
		if(curTime > 1){
			timer.Reset();
			setColor(((float)random() / ULONG_MAX), ((float)random() / ULONG_MAX), ((float)random() / ULONG_MAX));
		}
		else if(curTime > 0.75)
			setLight(false, true, false);
		else if(curTime > 0.5)
			setLight(false, false, true);
		else if(curTime > 0.25)
			setLight(false, true, false);
		else
			setLight(true, false, false);
		break;
		
	case kinect1:
		if(requestNextPattern) 
		{
			lightShow = kinect2;
			requestNextPattern = false;
		}
		setLight(true, true, true);
		{
	        float r, g, b;
	        hueToRGB(&r, &g, &b, rightArm);
	        r *= leftArm;
	        g *= leftArm;
	        b *= leftArm;
	        setColor(r, g, b);
		}

		break;
		
	case kinect2:
		if(requestNextPattern) 
		{
			lightShow = kinect3;
			requestNextPattern = false;
		}
		if(leftArm > .66)
			setLight(true, false, false);
		else if(leftArm > .33)
			setLight(false, true, false);
		else 
			setLight(false, false, true);
		if(rightArm > .66)
			setColor(1,0,0);
		else if(rightArm > .33)
			setColor(0,1,0);
		else
			setColor(0,0,1);
		break;

    case kinect3:
        if(requestNextPattern) 
        {
            lightShow = blink1;
            requestNextPattern = false;
        }
        if(leftArm > .66)
            setLight(true, false, false);
        else if(leftArm > .33)
            setLight(false, true, false);
        else 
            setLight(false, false, true);
        {
            float r, g, b;
            hueToRGB(&r, &g, &b, rightArm);
            setColor(r, g, b);
        }
        break;

	case blink1:
		if(requestNextPattern)
		{
			lightShow = binary;
			requestNextPattern = false;
		}
		
		if(alliance)
			setColor(1,0,0);
		else
			setColor(0,0,1);
		
		if(curTime > .5)
			timer.Reset();
		else if(curTime > .25)
			setLight(true, false, true);
		else 
			setLight(false, true, false);
		break;
		
	case binary:
		if(requestNextPattern)
		{
			lightShow = rider1;
			requestNextPattern = false;
		}
		if(alliance)
			setColor(1,0,0);
		else
			setColor(0,0,1);
		if(curTime > .5)
			++binaryCount;
		if(binaryCount > 7)
		{
			binaryCount = 0;
			timer.Reset();
		}
		setLight((bool)(binaryCount & 0x1),(bool)(binaryCount & 0x2),(bool)(binaryCount & 0x4));
		break;
        
    case rainbow:
    {
        // This one does not honor requestNextPattern on purpose - it just
        // stops after a certain time period as it is used to signal match
        // readiness
        requestNextPattern = false;
        const double rainbowTotalTime = 5.0;
        const int numSteps = 100;
        const double stepTime = rainbowTotalTime / numSteps;
        static double nextStepTime = 0;
        double now = timer.Get();
        if (now > rainbowTotalTime) {
            lightShow = ballPos;
            nextStepTime = 0;
	    } else if (now > nextStepTime){
	        float r,g,b;
	        now /= rainbowTotalTime;
	        hueToRGB(&r, &g, &b, now);
	        setColor(r, g, b);
	        setLight(true, true, true);
	        nextStepTime += stepTime;
        }
        break;
    }   

	default:
		lightShow = rider1;
		break;
	}
}

void Light::selectLightShow(LightShows pattern)
{
	lightShow = pattern;
	timer.Reset();
}

void Light::nextPattern()
{
	requestNextPattern = true;
}

void Light::showBallPos()
{
	lightShow = ballPos;
}

void Light::setColor(float red, float green, float blue)
{
	lightRed.UpdateDutyCycle(red);
	lightGreen.UpdateDutyCycle(green);
	lightBlue.UpdateDutyCycle(blue);
}

void Light::setLight(bool light1, bool light2, bool light3)
{
	lightSelect1.Set(light1);
	lightSelect2.Set(light2);
	lightSelect3.Set(light3);
}

void Light::setCoopertition(bool signal)
{
	signalRequest = signal;
}

void Light::setShooting(bool shootingStatus)
{
	shooting = shootingStatus;
}

void Light::setAlliance(bool color)
{
	alliance = color;
}
