#ifndef RPS_H
#define RPS_H

#include "WPILib.h"
#include "Camera.h"
//#include "ThunderDrive.h" - is this needed?

// lots of constants!
// how many times do we update position before we also update image?
/* TODO tune this value! */
#define UPDATES_BEFORE_NEW_IMAGE 10
// canned for compute time decrease
#define PI_180 3.14159265358979323/180
// convert g's to in/s/s
#define G_TO_IN_S2 9.81*39.37 
// width of drivetrain, inches TODO update this for 2012 bot!
#define DRIVEWIDTH 24
 // ticks per wheel revolution TODO check this
#define ONE_REV_TICKS 128
// circumference of wheel (in.) [check!]
#define ONE_REV_INCHES 8*3.14159265358979323
// inches traveled per encoder tick
#define INCHES_PER_TICK ONE_REV_INCHES/ONE_REV_TICKS
// weights for weighted average; see array weights[] below for more info
// TODO tune these values; MUST SUM TO 1.00 !!
#define WT_EE 1/6
#define WT_EG 1/6
#define WT_EA 1/6
#define WT_AE 1/6
#define WT_AG 1/6
#define WT_AA 1/6
// alternate weighting system: weighting angles and distances independently
// (weights[] will then contain these weights multiplied appropriately)
// TODO tune!; D's and A's must each sum to 1.00
#define WT_D_E 1/2
#define WT_D_A 1/2
#define WT_A_E 1/3
#define WT_A_G 1/3
#define WT_A_A 1/3
// and the selector between the two separate mechanisms
#define SPLIT_WEIGHTS 1
// when using camera data, weight to give camera pos vs. non-camera pos
// (non-camera pos gets weight 1-WT_CAM); TODO tune!!
#define WT_CAM 1/7
#define WT_NONCAM 1-WT_CAM
// field dimensions, in inches
#define MAX_X 54*12
#define MAX_Y 27*12


class RobotPositioningSystem {
public:
	// how do we get access to drivetrain encoders?
	RobotPositioningSystem(Camera *c); // constructor
	~RobotPositioningSystem();
	
	/** position conventions: y is parallel to alliance station wall, x is perpendicular 
	 * all xy values are in inches; (0,0) is the corner of the field at the inbounding station
	 * on our alliance station side; angle is trig standard (0 along x-axis, positive ccw) */
	
	// sets x and y position, and robot angle relative to field (more for init. purposes)
	void SetXYPosAngle(float x, float y, float theta);
	float GetXPos(); // returns x position
	float GetYPos(); // returns y position
	float GetAngle(); // returns robot angle relative to field
	// this should be called regularly and often:
	void UpdatePosition(); // tells the RPS to update its position calculation
	
	// do we need any functions to set things?
private:
	// Sensors
	ADXL345_I2C *accel;
	Camera *cam;
	Timer *timer;
		
	// Internal maintenance variables
	int update_counter; // how often we've updated our position without updating the image
	int lticks, rticks; // net ticks for the encoders
	float g_angle; // current gyro angle
	// current velocity (as calculated from accelerometer)
	float v_x, v_y; // (note: v_x is in direction of robot travel, v_y is perpendicular)
	double a_x, a_y; // acceleration (same as for velocity)
	float time; // elapsed time in seconds since last measurement
	/* more? */
	// temp arrays to hold changes in position; index is 0=ee, 1=eg, 2=ea, 3=ae, 4=ag, 5=aa
	// [the letters are distance-angle pairs, e=encoder,g=gyro,a=accelerometer]
	float dxs[6], dys[6];
	// weight to assign each of the above methods, based on reliability; puts #defined
	// constants above into more useful package
	float weights[6];
	// AND... the calculated position - if we've done it right
	float xpos, ypos;
	float angle; //robot angle relative to field
	
	// internal helper functions:
	// this one is put in a separate function to reduce code clutter
	void GetXYCamera(float* x, float* y);
	float GetGyroAngle(); // gets gyro angle, in radians
	double GetAX(); // gets acceleration in the x direction (in/s/s)
	double GetAY(); // gets acceleration in the y direction (in/s/s)
};

#endif
