#include "Storage.h"
#include "IOmap.h"
#include <math.h>

//Changed to make 2nd bot work for demo
//THROW THESE CHANGES AWAY AFTER THE 10 TURN POT IS INSTALLED
#if ROBOT_2
#define WINGS_POT_MIN 2.48
#define WINGS_POT_MAX 2.94
#else
#define WINGS_POT_MIN 2.48//4.18
#define WINGS_POT_MAX 3.00//4.66
#endif

//Middle 2.62
//.60 .602 Bridge
#define WINGS_HALF_UPPER_BOUND .30 //.27 center
#define WINGS_HALF_LOWER_BOUND .27

// Timeout for automovement of wings (in seconds)
#define WINGS_TIMEOUT 12.0

// Time wings stall at "off" when changing direction (to prevent current 
// spike that blows fuse) -- in seconds
#define BEATER_BAR_STALL_TIME 0.2f

#define WINGS_UP_SPEED   0.35f
#define WINGS_DOWN_SPEED -0.35f


Storage::Storage () {
	
	convBelt1 = new Relay(RELAY_CONV_STAGE_1_MOTOR);
	convBelt2 = new Relay(RELAY_CONV_STAGE_2_MOTOR);
	convBelt3 = new Relay(RELAY_CONV_STAGE_3_MOTOR);
#if ROBOT_2
	beaterBar = new Relay(RELAY_BEATER_BAR_MOTOR);
#else
	beaterBar = new Victor(RELAY_BEATER_BAR_MOTOR);
#endif

	bridgeStabilizer = new Victor(PWM_STABILIZER);
	bridgeStabilizerUnlocked = new DigitalInput(DIG_BRIDGE_STABILIZER);

	wingsPosition = new AnalogChannel(ANA_WING_POS_POT);
	wings = new ThunderCANJaguar(CANID_WINGS);
	wingsTimer = new Timer();
	
	countSensor = new DigitalInput(DIG_CONVEYOR_ENTRANCE);
	
	convPos1 = new AnalogChannel(ANA_CONV_STAGE_1);
	convPos2 = new AnalogChannel(ANA_CONV_STAGE_2);
	convPos3 = new AnalogChannel(ANA_CONV_STAGE_3);
	

	wingsDesiredPos = WINGS_STOP;
	beaterBarMode = BEATER_AUTO;
	beaterBarChangeTimer = new Timer();
	beaterBarChangeTimer->Start();
	beaterBarValue = Relay::kOff;
	beaterBarLastDir = Relay::kForward;
	conveyorBeltMode = CONV_AUTO;
	deliver = false;

	
	conv_sensors_broken = count_sensors_broken = false;
	wingSensorsBroken = false;
	sensor3_last = false;
	wingsManualSpeed = 0;
	wingsAutoDownSpeed = 0;
	wingsAtTop = false;
	wingsLastPos = 1.0;
}

Storage::~Storage() 
{
	delete convPos3;
	delete convPos2;
	delete convPos1;
	delete countSensor;
	delete wings;
	delete wingsPosition;
	delete wingsTimer;
	delete bridgeStabilizer;
	delete bridgeStabilizerUnlocked;
	delete beaterBar;
	delete beaterBarChangeTimer;
	delete convBelt3;
	delete convBelt2;
	delete convBelt1;
}

bool Storage::isReady() 
{
	return conv_sensors_broken || sensor3();
}

int Storage::processStorage() 
{
	// WINGS
    float wingsSpeed = 0;
    if (wingSensorsBroken || wingsDesiredPos == WINGS_STOP)
    {
        wingsSpeed = wingsManualSpeed;
        //printf("stopping wings\n");
    }
    else if (wingsTimer->Get() > WINGS_TIMEOUT)
    {
   		wingsSpeed = 0;
   		wingsTimer->Stop();
   		//printf("wings timing out\n");
    }
    else if (wingsDesiredPos == WINGS_UP)
    {
		if (getWingsPosition() == 1) {
			wingsSpeed = 0;
    		wingsTimer->Stop();
			wingsAtTop = true;
		} else {
			wingsSpeed = WINGS_UP_SPEED;
			wingsAtTop = false;
		}
	}
	else if (wingsDesiredPos == WINGS_DOWN)
	{
		if (getWingsPosition() == -1) {
    		wingsTimer->Stop();
		} else {
		    float pos = wingsPos();
            wingsSpeed = wingsAutoDownSpeed;

            if (fabs(pos - wingsLastPos) < 0.05) {
		        wingsAutoDownSpeed -= 0.01;
		        if (wingsAutoDownSpeed < -1.0)
		            wingsAutoDownSpeed = -1.0;
            }
		    wingsLastPos = pos;
		}
	}
	else if(wingsDesiredPos == WINGS_HALF)
	{
		if (getWingsPosition() == 2) {
    		wingsTimer->Stop();
		}
		else {
			if(wingsPos() >= WINGS_HALF_UPPER_BOUND)
				wingsSpeed = WINGS_DOWN_SPEED;
			else if(wingsPos() <= WINGS_HALF_LOWER_BOUND)
				wingsSpeed = WINGS_UP_SPEED;
		}
	}
	else
	{
		wingsSpeed = 0;
	}
    wings->Set(wingsSpeed);

    // CONVEYOR BELTS
	if (conveyorBeltMode == CONV_AUTO)
	{
		if (!conv_sensors_broken)
		{
			
			// SIMPLE CODE!
			
			// If the current area is empty or the next is, move forward. Only stop
			// when this and the next areas are full. PROBLEM: Ball1 in Zone1(end),
			// Ball2 in Zone2(intermediate). The code would advance both, so both balls
			// could be on the Zone2 conveyor. This may not be bad. Only impacts
			// counts->beater bar
			
			if (!sensor1() || !sensor2()) convBelt1Move(1);
			else convBelt1Move(0);

			if (!sensor2() || !sensor3()) convBelt2Move(1);
			else convBelt2Move(0);
		
			// notice that deliver should be true only after sensor3 = true;
			if (!sensor3() || deliver) convBelt3Move(1);
			else convBelt3Move(0);
			
			if (deliver && !sensor3() && sensor3_last) deliver = false;
			sensor3_last = sensor3();

		} 
		
		else
		{
		    convBelt1Move(0);
            convBelt2Move(0);
            convBelt3Move(0);
		}
	}
	else if (conveyorBeltMode == CONV_IN)
	{
        convBelt1Move(1);
        convBelt2Move(1);
        convBelt3Move(1);
	}
	else if (conveyorBeltMode == CONV_OUT)
	{
        convBelt1Move(-1);
        convBelt2Move(-1);
        convBelt3Move(-1);
	}
	else if (conveyorBeltMode == CONV_OUT_1)
	{
        convBelt1Move(-1);
        convBelt2Move(0);
        convBelt3Move(0);		
	}
	else
	{
	 convBelt1Move(0);
     convBelt2Move(0);
     convBelt3Move(0);
     }
	// BEATER BAR
	if (beaterBarMode == BEATER_AUTO)
	{
		if ( !(sensor1() && sensor2() && sensor3())   &&   !sensor0() )
		{
			beaterBarSet(Relay::kForward);
		}
		else if (sensor1() && sensor2() && sensor3())
		{
			beaterBarSet(Relay::kReverse);
		}
		else if (sensor0()) 
		{
			beaterBarSet(Relay::kOff);
		}
	}
	else if (beaterBarMode == BEATER_IN)
	{
		beaterBarSet(Relay::kForward);
	}
	else if (beaterBarMode == BEATER_OUT)
	{
		beaterBarSet(Relay::kReverse);
	}
	else
	{
	    // Stop mode
	    beaterBarSet(Relay::kOff);
	}
	return getBallCount();
}


int Storage::getBallCount() 
{
	// QUESTION; why have three variables, effectively functionally indistinguishable,
	// when one suffices? TODO
    int pos1Count;
    int pos2Count;
    int pos3Count;
    
    if (sensor1())
    {
        pos1Count = 1;
    }
    else
    {
        pos1Count = 0;
    }
        
    if (sensor2())
    {
        pos2Count = 1;
     }
    else
    {
        pos2Count = 0;
    }
        
    if (sensor3())
    {
        pos3Count = 1;
    }
    else
    {
        pos3Count = 0;
    }
    
    return (pos1Count + pos2Count + pos3Count);
}

void Storage::deliverBall() 
{
	if (!conv_sensors_broken) 	
	{
		deliver = true;
	}
}

bool Storage::deliverInProgress()
{
	return deliver;
}

void Storage::cancelDeliverBall() 
{
    deliver = false;
}




//Wings------------------------------------------------------------------------------------------------------

int Storage::getWingsPosition()
{
	if (!wingSensorsBroken && wingsPos() <= 0.05)
	{
		//printf("returing wings at down\n");
	return -1;
	
	}
	
	else if (!wingSensorsBroken && wingsPos() >= .95)
	{
		//printf("returning wings at up\n")l
	return 1;
	}
	
	else if(!wingSensorsBroken && wingsPos() >= WINGS_HALF_LOWER_BOUND && wingsPos() <= WINGS_HALF_UPPER_BOUND)
	{
		return 2;
	}
	else
	{
	return 0;	
	}
}

void Storage::wingsMove(int dir)
{
	
	if(dir == -1)
        wingsManualSpeed = WINGS_DOWN_SPEED;
	else if(dir == 1)
	    wingsManualSpeed = WINGS_UP_SPEED;
	else
	    wingsManualSpeed = 0;
	
	wingsDesiredPos = WINGS_STOP;
}

void Storage::setWingsPosition(WingsPosition pos) 
{
    if (wingSensorsBroken)
        // Don't do auto movements if broken
        return;
	
    if (pos == WINGS_DOWN && wingsDesiredPos != WINGS_DOWN)
	{	
        wingsTimer->Reset();
        wingsTimer->Start();
        wingsDesiredPos = WINGS_DOWN;
        wingsAutoDownSpeed = WINGS_DOWN_SPEED;
        wingsLastPos = wingsPos();
        //printf("desired position: down\n");
	}
	
	if (pos == WINGS_UP && wingsDesiredPos != WINGS_UP)
	{
        wingsTimer->Reset();
        wingsTimer->Start();	    
	    wingsDesiredPos = WINGS_UP;
	    //printf("desired position: up\n");
	}

	if (pos == WINGS_HALF && wingsDesiredPos != WINGS_HALF)
	{
		wingsTimer->Reset();
		wingsTimer->Start();
		wingsDesiredPos = WINGS_HALF;
		//printf("desired position: half\n");
	}
	
	wingsManualSpeed = 0;
}


void Storage::wingsStop()
{
	
	wingsDesiredPos = WINGS_STOP;
	wings->Set(0);
	wingsManualSpeed = 0;
}


float Storage::wingsPos()
{
	float wingsPosNumber;
	
	wingsPosNumber = wingsPosition->GetVoltage();
	
	wingsPosNumber = (wingsPosNumber - WINGS_POT_MIN) / (WINGS_POT_MAX - WINGS_POT_MIN);
	
	//need more information to do this
	
	return wingsPosNumber;
}

bool Storage::atWingsPosition()
{
    if (wingsDesiredPos == WINGS_UP && (getWingsPosition() == 1 || wingsTimer->Get() > WINGS_TIMEOUT))
    {
        return true;
    }
    
    else if (wingsDesiredPos == WINGS_DOWN && (getWingsPosition() == -1 || wingsTimer->Get() > WINGS_TIMEOUT))
    {
        return true;
    }
    
    else if (wingsDesiredPos == WINGS_HALF && (getWingsPosition() == 2 || wingsTimer->Get() > WINGS_TIMEOUT))
    {
        return true;
    }
    
    else
    {
        return false;
    }
}

void Storage::setWingsBroken(bool broken)
{
	if (!wingSensorsBroken && broken) {
	    wingsMove(0);
	}
	wingSensorsBroken = broken;
}

bool Storage::areWingsBroken()
{
	return wingSensorsBroken;
}

float Storage::wingsRawValue()
{
    return wingsPosition->GetVoltage();
}
//Wings------------------------------------------------------------------------------------------------------

void Storage::setConvBroken(bool broken)
{
	conv_sensors_broken = broken;
}

bool Storage::isConvBroken()
{
	return conv_sensors_broken;
}

void Storage::setCountBroken(bool broken)
{
	count_sensors_broken = broken;
}

bool Storage::isCountBroken()
{
	return count_sensors_broken;
}

void Storage::beaterBarMove(int dir) 
{
	if (dir == 1)
	{
		beaterBarMode = BEATER_IN;
	}
	else if (dir == -1)
	{
		beaterBarMode = BEATER_OUT;
	}
	else if (dir == 0)
	{
		beaterBarMode = BEATER_AUTO;
	}
	else
	{
	    beaterBarMode = BEATER_STOP;
	}   
} 

int Storage::getBeaterBarMode()
{
    if (beaterBarMode == BEATER_AUTO)
    {
        return 0;
    }
    else if (beaterBarMode == BEATER_OUT)
    {
        return -1;
    }
    else if (beaterBarMode == BEATER_IN)
    {
        return 1;
    }
    else
    {
        return -99;
    }
}

void Storage::beaterBarSet(Relay::Value val)
{
    Relay::Value output = val;
    if (val != Relay::kOff && beaterBarLastDir != val) {
        // New direction - different from last driven direction
        // Stall for some time to avoid current spikes that blow fuse
        if (beaterBarValue != Relay::kOff) {
           // Dir change must go through stop
           // timer below will reset in this case
           output = Relay::kOff;
        } else if (beaterBarChangeTimer->Get() > BEATER_BAR_STALL_TIME) {
           // ok, stopped long enough
           beaterBarLastDir = val;
        } else {
           // waiting for stall time
          output = Relay::kOff;
        }
    }
    if (beaterBarValue != Relay::kOff && output == Relay::kOff)
        // Stopping after being not stopped
        // Track how long we've been stopped
        beaterBarChangeTimer->Reset();
#if ROBOT_2
    beaterBar->Set(output);
#else
    if(output == Relay::kForward)		//changed code to operate a Victor instead of a relay
    	beaterBar->Set(-1);
    else if(output == Relay::kReverse)
    	beaterBar->Set(1);
    else
    	beaterBar->Set(0);
#endif
    //beaterBar->Set(output);
    beaterBarValue = output;
}

void Storage::setBridgeStabilizer(float speed)
{
	if (speed > 0 && bridgeStabilizerUnlocked->Get())
		speed = 0;
	bridgeStabilizer->Set(speed);
}

bool Storage::isBridgeStabilizerLocked()
{
	return !bridgeStabilizerUnlocked->Get();
}

void Storage::convBelt1Move(int dir) 
{
	if (dir == 1)
		{
			convBelt1->Set(Relay::kForward);
		}
	else if (dir == -1)
		{
			convBelt1->Set(Relay::kReverse);
		}
	else
		{
			convBelt1->Set(Relay::kOff);
		}	
}
void Storage::convBelt2Move(int dir) 
{
	if (dir == 1)
		{
			convBelt2->Set(Relay::kForward);
		}
		else if (dir == -1)
		{
			convBelt2->Set(Relay::kReverse);
		}
		else
		{
			convBelt2->Set(Relay::kOff);
		}
}
void Storage::convBelt3Move(int dir) 
{
	if (dir == 1)
	{
		convBelt3->Set(Relay::kForward);
	}
	else if (dir == -1)
	{
		convBelt3->Set(Relay::kReverse);
	}
	else
	{
		convBelt3->Set(Relay::kOff);
	}
}

void Storage::manualConvMove(int dir)
{
	if (dir == 1)
	{
		conveyorBeltMode = CONV_IN;
	}
	else if (dir == -1)
	{
		conveyorBeltMode = CONV_OUT;
	}
	else if (dir == 0)
	{
		conveyorBeltMode = CONV_AUTO;
	}
	else if (dir == -2)
	{
		conveyorBeltMode = CONV_OUT_1;
	}
	else
	{
		conveyorBeltMode = CONV_STOP;
	}
}

bool Storage::getSensorState(int sensor)
{
    switch (sensor) 
    {
    case 0:
        return sensor0();
    case 1:
        return sensor1();
    case 2:
        return sensor2();
    case 3:
        return sensor3();
    default:
        return false;
    }
}

float Storage::getSensorVoltage(int sensor)
{
    switch (sensor)
    {
    case 0:
            if (countSensor->Get() == 0)
            {
                return 0;
            }
            else
            {
                return 5;
            }
            
    case 1:
            return convPos1->GetVoltage();
    case 2:
            return convPos2->GetVoltage();
    case 3:
            return convPos3->GetVoltage();
    default:
        return 9001;
    }
}

bool Storage::sensor0()
{
	if (countSensor->Get() == 0)
	{
		return false;
	}
	else
	{
		return true;
	}
}

bool Storage::sensor1()
{
	if (convPos1->GetVoltage() > 3.5)
	{
		return true;
	}
	else
	{
		return false;
	}
}

bool Storage::sensor2() 
{
	if (convPos2->GetVoltage() > 3.5)
	{
		return true;
	}
	else
	{
		return false;
	}
}


bool Storage::sensor3()
{
	if (convPos3->GetVoltage() > 2.6)
	{
		return true;
	}
	else
	{
		return false;
	}
}
