#include "ThunderAuto.h"

#define USE_GYRO 0

#define BRIDGE_CLIMBING 24				//The amount needed to drive up a bridge in inches
#define ONE_FIRST_DRIVE 50				//The driving to the cooperation bridge in position one in inches
#define AUTO_DRIVE_SPEED  .4			//The speed at which we drive in any direction

#define POS_ONE_CAMERA_ANGLE 45			//The camera angle of position 1
#define POS_TWO_CAMERA_ANGLE 55			//The camera angle of position 2
#define POS_THREE_CAMERA_ANGLE 55		//The camera angle of position 3
#define COOP_BRIDGE_CAMERA_ANGLE 30		//The camera angle at the cooperation bridge
#define ALLY_BRIDGE_CAMERA_ANGLE 30	 	//The camera angle at the alliance bridge

#if ROBOT_2
#define POS_ONE_TURRET_ANGLE -194 		//The Turret angle at position 1
#define POS_TWO_TURRET_ANGLE  110		//The Turret angle at position 2
#define POS_THREE_TURRET_ANGLE 23		//The Turret angle at position 3
#define COOP_BRIDGE_TURRET_ANGLE -194	//The Turret angle at the cooperation bridge
#define ALLY_BRIDGE_TURRET_ANGLE -160	//The turret angle at the alliance bridge
#else
#define POS_ONE_TURRET_ANGLE 0		//The Turret angle at position 1
#define POS_TWO_TURRET_ANGLE 0		//The Turret angle at position 2
//#define POS_ONE_TURRET_ANGLE -190		//The Turret angle at position 1
//#define POS_TWO_TURRET_ANGLE  110		//The Turret angle at position 2
#define POS_THREE_TURRET_ANGLE 15	//The Turret angle at position 3 //MAB need to tune this
#define COOP_BRIDGE_TURRET_ANGLE -189	//The Turret angle at the cooperation bridge
#define ALLY_BRIDGE_TURRET_ANGLE -160	//The turret angle at the alliance bridge
#endif

#if ROBOT_2
#define POS_ONE_TURRET_SPEED_TOP .85			//The shooter speed at Position1
#define POS_ONE_TURRET_SPEED_BOTTOM .85			//The shooter speed at Position1
#define POS_TWO_TURRET_SPEED_TOP .7			//the shooter speed at Position2
#define POS_TWO_TURRET_SPEED_BOTTOM .7			//the shooter speed at Position2
#define POS_THREE_TURRET_SPEED_TOP .65		//the shooter speed at Position3
#define POS_THREE_TURRET_SPEED_BOTTOM .65		//the shooter speed at Position3
#else
#define POS_ONE_TURRET_SPEED_TOP .95			//The shooter speed at Position1
#define POS_ONE_TURRET_SPEED_BOTTOM .95			//The shooter speed at Position1
#define POS_TWO_TURRET_SPEED_TOP .7			//the shooter speed at Position2
#define POS_TWO_TURRET_SPEED_BOTTOM .7			//the shooter speed at Position2
#define POS_THREE_TURRET_SPEED_TOP .70		//the shooter speed at Position3
#define POS_THREE_TURRET_SPEED_BOTTOM .70		//the shooter speed at Position3
#endif
#define COOP_BRIDGE_TURRET_SPEED_TOP 1		//the shooter speed at the cooperation bridge
#define COOP_BRIDGE_TURRET_SPEED_BOTTOM 1		//the shooter speed at the cooperation bridge

#define ERROR_MARGIN 25 				//The number of degrees of error to allow when driving


ThunderAuto::ThunderAuto(ThunderDrive *drive, SmartShooter *shooter, Storage *storage)
{
	ThunderDriveCode = drive;
	Thunderstorage = storage;
	Thundershootercode = shooter;
	
	initial_state = ThunderAuto::finishin;
	posone_state = ThunderAuto::one_done;
	postwo_state = ThunderAuto::two_done;
	autoshot_state = ThunderAuto::finish;
	//1507
	pos1507_state = ThunderAuto::t1507_done;
	//1507
	
	Moving_Wings = false;
}

ThunderAuto::~ThunderAuto() {
	
}

void ThunderAuto::updateAngle(float desiredAngle)
{
	if (!USE_GYRO)
	{
		return;
	}
	
	float currentAngle = ThunderDriveCode->getGyro();

	float newAngle = currentAngle + desiredAngle;
	if (newAngle >= currentAngle)
	{
		minAngle = currentAngle - ERROR_MARGIN;
		maxAngle = newAngle + ERROR_MARGIN;
	}
	else
	{
		minAngle = newAngle - ERROR_MARGIN;
		maxAngle = currentAngle + ERROR_MARGIN;
	}

}

bool ThunderAuto::checkCourse()
{
	if (!USE_GYRO)
	{
		return true;
	}
	
	float currentAngle = ThunderDriveCode->getGyro();
	if ((currentAngle >= minAngle) && (currentAngle <= maxAngle))
	{
		return true;
	}
	else
	{
		printf("off course\n");
		endAuto(false);
		return false;
	}
}


// initialize autonomous (call it once)
void ThunderAuto::InitializeAuto(int position, int steps, int delay)
{
	 Robopos = position;
	 Step_Count = steps;
	 Delay_Time = delay;
	 delayTime = delay;
	 Moving_Wings = false;
	 printf("Robopos:%d   Step_Count:%d   Delay_Time:%d\n", Robopos, Step_Count, Delay_Time);
	 
	if (Robopos == 0)
	{
		printf("InitializeAuto: do nothing\n");
		endAuto(true);
		return;
	}
	//We don't want automatic backspin because automode already accounts for this
	Thundershootercode->setUseBackspin(false);
	
	if (delay == 0)
	{
		// set state to the one after the delay
		initial_state = ThunderAuto::inshoot_one;
	}
	else
	{
		 //set to delay state
		initial_state = ThunderAuto::delay;
	}

	//1507
	if ((Robopos == 3) && (Step_Count == 3))
	{
		printf("InitializeAuto: init Process1507\n");
		initial_state = ThunderAuto::finishin;//hack to minimize cut-n-paste/merge
		pos1507_state = ThunderAuto::t1507_delay;
		begin_time = GetTime();
	}
	//1507
	
	if ((Robopos == 3) && (Step_Count == 4))
	{
		printf("InitializeAuto: init 3 4\n");
		initial_state = ThunderAuto::inauto_check;//Skip Shooting
		Robopos = 1;//We're running pos 1's auto mode, just shorter
		Step_Count = 3;
		InitializePosition_one();
	}

	ProcessAuto();
}

// process all auto steps (call it for every loop in autonomous)
bool ThunderAuto::ProcessAuto()
{
	//1507
	if ((Robopos == 3) && (Step_Count == 3))
	{
		if (Process1507())
		{
			printf("ProcessAuto: Position 1507 is done\n");
			endAuto(true);
			return (true);
		}
		return (false);
	}
	//1507
	
	// put wings half way if doing 'any' auto; even 0 0 X
	if ((Robopos != 0) || (Step_Count != 0) || (delayTime != 0.0))
	{
		// don't override internal wing movement
		if (!Moving_Wings) {
			Thunderstorage->setWingsPosition(Storage::WINGS_HALF);
		}
	}

	switch (initial_state)
	{
		case ThunderAuto::delay:
		{	
			begin_time = GetTime();
			initial_state = ThunderAuto::delay_check;
			printf("ProcessAuto: Waiting *delay*\n");
			break;
		}
	
		case ThunderAuto::delay_check:
		{
			double curTime = GetTime();
			if (curTime >= (begin_time + (double)delayTime))
			{
				initial_state = ThunderAuto::inshoot_one;
				printf("ProcessAuto: Delay is done *delay_check*\n");
			}                       
			break;
		}
		
		case ThunderAuto::inshoot_one:
		{
			if (Step_Count < 1)
			{
				printf("ProcessAuto: Not enough steps, ending Autonomous *inshoot_one*\n");
				endAuto(true);
				return (true);
			}
			switch (Robopos)
			{
				default:	// do nothing
					printf ("ProcessAuto: Invalid Robot Position *inshoot_one*\n");
					endAuto(true);
					return (true);
					
				case 1:
				{
					printf ("ProcessAuto: Initializing position one shooting *inshoot_one*\n");
					Initialize_Shooter(2, false, 9001, POS_ONE_CAMERA_ANGLE, POS_ONE_TURRET_ANGLE, POS_ONE_TURRET_SPEED_TOP, POS_ONE_TURRET_SPEED_BOTTOM);						
					break;	
				}
					
				case 2:
				{
					printf ("ProcessAuto: Initializing position two shooting *inshoot_one*\n");
					Initialize_Shooter(2, false, 9001, POS_TWO_CAMERA_ANGLE, POS_TWO_TURRET_ANGLE, POS_TWO_TURRET_SPEED_TOP, POS_TWO_TURRET_SPEED_BOTTOM);
					break;
				}

				case 3:
				{
					printf ("ProcessAuto: Initializing position three shooting *inshoot_one*\n");
					Initialize_Shooter(2, false, 9001, POS_THREE_CAMERA_ANGLE, POS_THREE_TURRET_ANGLE, POS_THREE_TURRET_SPEED_TOP, POS_THREE_TURRET_SPEED_BOTTOM);
					break;
				}
			}
			
			initial_state = ThunderAuto::inshoot_check;
			break;
		}
			
		case ThunderAuto::inshoot_check:
		{
			if (Autoshots())
			{
				if (Step_Count < 2)
				{
					printf("ProcessAuto: No more steps, ending *inshoot_check*\n");
					endAuto(true);
					return(true);
				}
				
				switch (Robopos)
				{
					default:
					{
						printf("ProcessAuto: Invalid Position *inshoot_check*\n");
						endAuto(true);
						return(true);
					}
					case 1:
					{
						printf("ProcessAuto: Starting Position One *inshoot_check*\n");
						InitializePosition_one();
						break;
					}
				}
				initial_state = inauto_check;
			}
			break;
		}
		
		case ThunderAuto::inauto_check:
		{
			switch (Robopos)
			{
				default:
				{
					printf("ProcessAuto: Invalid Position *inauto_check*\n");
					endAuto(true);
					return(true);
				}
				case 1:
				{
					if (Position_one())
					{
						printf("ProcessAuto: Position one is done *inauto_check*\n");
						endAuto(true);
						return (true);
					}
					break;
				}
			}
			break;
		}
		
		case ThunderAuto::finishin:
		{
			// NO PRINTF
			return (true);
		}
	}
	return (false);
}

void ThunderAuto::Initialize_Shooter(int max_balls, bool ask_robot_has_balls, int wait_time, int cam_angle, int tur_angle, double tur_speed_top, double tur_speed_bottom)
{
	//initialized whenever autoshots is called
	autoshot_state = ThunderAuto::Camturn_check;
	Max_Balls = max_balls;
	Ask_Robot_Has_Balls = ask_robot_has_balls;
	Wait_Time = wait_time;
	Thunderstorage->manualConvMove(0);
	Thunderstorage->beaterBarMove(0);
	Thundershootercode->tiltCamera(cam_angle);	
	Thundershootercode->setTurretAngle(Thundershootercode->getTurretAngle());
	Thundershootercode->setNormalizedSpeed(tur_speed_top, tur_speed_bottom);
	updateAngle (0);

	begin_time = GetTime();
	Autoshots();
}

bool ThunderAuto::Autoshots()
{
	switch (autoshot_state)
	{
		case ThunderAuto::Camturn_check:
		{
			if (Thundershootercode->turretFinished()==true)
			{
				printf("Autoshots: Turret ready *camturn_check*\n");
				autoshot_state = ThunderAuto::ball_check;
			}
			break;
		}
		
		case ThunderAuto::ball_check:
		{	
			// if we shot all requested balls, then we are done
			if (Max_Balls <= 0)
			{
				printf("Autoshots: shot all balls\n");
				autoshot_state = finish;
				return (true);
			}

			// if we are not waiting for the robot to tell us it has		
			// balls OR the robot ball counter is broken, then we
			// just go ahead and shoot (i.e. nothing to wait for)
			if (!Ask_Robot_Has_Balls || Thunderstorage->isConvBroken())
			{
				printf("Autoshots: shoot\n");
				Thundershootercode->shoot();
				begin_time_shoot = GetTime();
				autoshot_state = shoot_check_wait;
				break;
			}
		
			// if the robot has a ball, then shoot it
			if (Thunderstorage->getBallCount() > 0)
			{
				printf("Autoshots: have more balls\n");
				Thundershootercode->shoot();
				begin_time_shoot = GetTime();
				autoshot_state = shoot_check_wait;
				break;
			}
		
			// if time has expired, then done shooting
			// NOTE: this is a total timeout, not a per ball timeout
			double curTime = GetTime();
			if (curTime >= (begin_time + (double)Wait_Time))
			{
				printf("Autoshots: timeout waiting for balls\n");
				autoshot_state = finish;
				return (true);
			}
			
			// continue to wait for next ball
			break;
		}	
		
		case ThunderAuto::shoot_check_wait:
		{	
			double curTime = GetTime();
			if (curTime >= (begin_time_shoot + (double)2.0))
			{
				printf("Autoshots: timeout waiting to check shoot\n");
				autoshot_state = shoot_check;
				break;
			}
			Thundershootercode->shoot();
			break;
		}
		
		case ThunderAuto::shoot_check:
		{	
			if (Thundershootercode->shootFinished())
			{
				printf("Autoshots: shot a ball\n");
				Max_Balls--;
				autoshot_state = ball_check;
			}
			break;
		}
		
		case ThunderAuto::finish:
		{
			// NO PRINTF
			return (true);
		}
	}

	return (!checkCourse());
}

void ThunderAuto::InitializePosition_one()
{
	posone_state = one_drive_one;
	Position_one();
}

bool ThunderAuto::Position_one()
{
	switch (posone_state)
	{		
		case ThunderAuto::one_drive_one:
		{
			printf("Position_one: drive\n");
			ThunderDriveCode ->autonomousMove(ONE_FIRST_DRIVE, AUTO_DRIVE_SPEED);
			posone_state = ThunderAuto::one_drive_one_check;
			updateAngle(0);
			break;
		}	
	
		case ThunderAuto::one_drive_one_check:
		{	
			if (ThunderDriveCode->doneYet() == true)
			{
				if ((Step_Count < 3) || Thunderstorage->areWingsBroken())
				{
					printf("Position_one: no more steps or wings broken\n");
					endAuto(true);
					return(true);
				}

				printf("Position_one: put wings down\n");
				Moving_Wings = true; // leave true for duration of auto
				Thunderstorage->setWingsPosition(Storage::WINGS_DOWN);
				posone_state = ThunderAuto::one_lower_actuator_check;
			}
			break;
		}		
	
		case ThunderAuto::one_lower_actuator_check:
		{	
			if (Thunderstorage->atWingsPosition() == true)
			{
				printf("Position_one: wings are down\n");
				if (Step_Count >= 4)
				{
					//have more steps, so don't wait forever
					ThunderAuto::Initialize_Shooter(16,true, 5, COOP_BRIDGE_CAMERA_ANGLE, COOP_BRIDGE_TURRET_ANGLE, COOP_BRIDGE_TURRET_SPEED_TOP, COOP_BRIDGE_TURRET_SPEED_BOTTOM);
				}
				else
				{
					ThunderAuto::Initialize_Shooter(16,true, 9001, COOP_BRIDGE_CAMERA_ANGLE, COOP_BRIDGE_TURRET_ANGLE, COOP_BRIDGE_TURRET_SPEED_TOP, COOP_BRIDGE_TURRET_SPEED_BOTTOM);
				}
				posone_state = ThunderAuto::one_shoot_two_check;
			}
			break;
		}
		
		case ThunderAuto::one_shoot_two_check:
		{	
			if (Autoshots())
			{
				if (Step_Count < 4)
				{
					printf("Position_one: not 4 steps, ending\n");
					endAuto(true);
					return(true);
				}

				printf("Position_one: drive up bridge\n");
				ThunderDriveCode->autonomousMove(BRIDGE_CLIMBING, AUTO_DRIVE_SPEED);
				posone_state = ThunderAuto::one_bridge_check;
				updateAngle(0);
			}																						
			break;
		}	
	
		case ThunderAuto::one_bridge_check:
		{	
			if (ThunderDriveCode->doneYet()== true)
			{
				printf("Position_one: done driving up bridge\n");
				endAuto(true);
				return (true);
			}
			break;
		}
		
		case ThunderAuto::one_done:
		{
			return (true);
		}
	}	

	return (!checkCourse());
}			


void ThunderAuto::endAuto(bool good_boy) 
{
	if (good_boy)
	{
		initial_state = ThunderAuto::finishin;
		posone_state = ThunderAuto::one_done;
		postwo_state = ThunderAuto::two_done;
		autoshot_state = ThunderAuto::finish;
		printf ("endAuto: We are done\n");
	}
	else
	{
		initial_state = ThunderAuto::finishin;
		posone_state = ThunderAuto::one_done;
		postwo_state = ThunderAuto::two_done;
		autoshot_state = ThunderAuto::finish;
		printf ("endAuto: An error has occured\n");
	}
	//1507
	pos1507_state = ThunderAuto::t1507_done;
	Thunderstorage->beaterBarMove(-100);
	Thunderstorage->manualConvMove(-100);
	//1507
	ThunderDriveCode->stopAction(); 
	Thunderstorage->wingsStop();
	Thundershootercode->shootStop();
	Thundershootercode->stopTurret();
}

//1507
bool ThunderAuto::Process1507()
{
	switch (pos1507_state)
	{
		case ThunderAuto::t1507_delay:	// wait before deliverying first ball (based on control's value)
		{
			double curTime = GetTime();
			if (curTime >= (begin_time + (double)delayTime))
			{
				pos1507_state = ThunderAuto::t1507_ball1;
				begin_time = GetTime();
				Thunderstorage->beaterBarMove(-1);
				Thunderstorage->manualConvMove(-2);
				printf("Process1507: start to move ball 1\n");
			}                       
			break;
		}
		
		case ThunderAuto::t1507_ball1:	// move out ball 1 until we need to move ball 2
		{
			double curTime = GetTime();
			if (curTime >= (begin_time + (double)4))// delay is based on what 1507 wants
			{
				pos1507_state = ThunderAuto::t1507_ball2;
				begin_time = GetTime();
				Thunderstorage->beaterBarMove(-1);
				Thunderstorage->manualConvMove(-1);
				printf("Process1507: start to move ball 2\n");
			}                       
			break;
		}
		
		case ThunderAuto::t1507_ball2:	// move out ball 2
		{
			double curTime = GetTime();
			if (curTime >= (begin_time + (double)3))// delay is how long it takes to eject ball
			{
				printf("Process1507: all done\n");
				endAuto(true);
				return (true);
			}                       
			break;
		}
				
		case ThunderAuto::t1507_done:
		{
			// NO PRINTF
			return (true);
		}
	}
	return (false);
}
//1507
