#include "UDPLogger.h"
#include <inetLib.h>
#include <stdio.h>
#include <string.h>

#define ENABLE_UDP_LOGGER


UDPLogger::UDPLogger(bool print) : fd(0), stdoutprint(print), enabled(true) {
}

UDPLogger::~UDPLogger() {
	if (fd)
		close(fd);
}

void UDPLogger::init(int port) {
#ifdef ENABLE_UDP_LOGGER
	memset(&saddr, 0, sizeof(saddr));
	saddr.sin_family = AF_INET;
	saddr.sin_port = htons(port);
	
	if (inet_aton("10.15.11.10", &saddr.sin_addr) == ERROR) {
		printf("UDPLogger: Bad inet_aton\n");
		return;
	}

	int tmpfd = socket(AF_INET, SOCK_DGRAM, 0);
	if (tmpfd == ERROR) {
		printf("UDPLogger: Bad socket\n");
		return;
	}	
	fd = tmpfd;
#endif
}

void UDPLogger::log(const char *fmt, ...)
{
    if (enabled) {
    	if (fd) {
    		va_list args;
    		va_start(args, fmt);
    		vsprintf(buf, fmt, args);
    		va_end(args);
    		if (sendto(fd, buf, strlen(buf), 0, (sockaddr *)&saddr, sizeof(saddr)) == ERROR) {
    			printf("UDPLogger: sendto failed\n");
    		}
    	}
    	if (stdoutprint) {
        	va_list args;
        	va_start(args, fmt);
        	vprintf(fmt, args);
        	va_end(args);
    	}
    }
}

void UDPLogger::setEnabled(bool en) {
    enabled = en;
}

LoggingPIDOutput::LoggingPIDOutput(PIDOutput *out, UDPLogger *logger, MonoEncoder *encoder) : 
                                       controller(NULL),
                                       out(out),
                                       logger(logger),
                                       encoder(encoder)
{
}

LoggingPIDOutput::~LoggingPIDOutput() {
}

void LoggingPIDOutput::setController(PIDController *controller) {
    this->controller = controller;
}

void LoggingPIDOutput::PIDWrite(float output) {
	if (controller != NULL) {
		if (output == 0) {
			logger->log("0,0,0,0\n"); //This is to prevent locking in a critical section
		} else {
			//logger->log("%f,%f,%f\n", (double)controller->GetError(), (double)output, (double)controller->GetSetpoint());
			logger->log("%f,%f,%f,%f\n", (double)controller->GetError(), (double)output, (double)encoder->getLastSpeed(), (double)controller->GetSetpoint());
		}
	}
    out->PIDWrite(output);
}
