#include "WPILib.h"
#include "Controls.h"
#include "Light.h"
#include "SmartShooter.h"
#include "Storage.h"
#include "ThunderAuto.h"
#include "ThunderDrive.h"
#include "DSmap.h"

class ThunderBot : public SimpleRobot
{
	Storage* thunderStorage;
	SmartShooter* thunderShooter;
	ThunderDrive* thunderDrive;
	ThunderAuto* thunderAuto;
	Light* thunderLight;
	Controls* thunderControls;
	KinectStick* leftKinect;
	KinectStick* rightKinect;
	Joystick* leftJoystick;
	Joystick* rightJoystick;
	
public:
	ThunderBot(void)
	{
		leftKinect = new KinectStick(1);
		rightKinect = new KinectStick(2);
		leftJoystick = new Joystick(DS_JOY_PORT_LEFT);
		rightJoystick = new Joystick(DS_JOY_PORT_RIGHT);
		thunderStorage = new Storage();
        thunderLight = new Light(thunderStorage,leftJoystick, rightJoystick);
        thunderLight->showStartupWarning(true);
		thunderShooter = new SmartShooter(thunderStorage);
		thunderDrive = ThunderDrive::getInstance();
		thunderAuto = new ThunderAuto(thunderDrive, thunderShooter, thunderStorage);
		thunderControls = new Controls(thunderShooter, thunderStorage, thunderDrive, thunderLight, leftJoystick, rightJoystick);
        thunderLight->showStartupWarning(false);
        thunderLight->selectLightShow(Light::rainbow);
	}
	
	~ThunderBot()
	{
	    delete thunderControls;
	    delete thunderLight;
        delete thunderAuto;
	    delete thunderShooter;
        delete thunderStorage;
        delete thunderDrive;
	}
	
	void Disabled(void) {
		while (IsDisabled())
		{
			thunderControls->processBrokenSwitches();
			thunderControls->processDebug();
			thunderControls->processDashboard();
			thunderControls->processLightControls();
			thunderLight->processLightShow();			
			Wait(0.005);
		}
	}

	void Autonomous(void)
	{
		int startPos;
		int steps;
		int delay;
		thunderControls->getAutoValues(&startPos, &steps, &delay);
		thunderAuto->InitializeAuto(startPos, steps, delay);
//		thunderShooter->requestZero();
		while(IsAutonomous() && IsEnabled())
		{
			thunderControls->processBrokenSwitches();
			thunderAuto->ProcessAuto();
			thunderControls->processDashboard();
			thunderShooter->processShooter();
			thunderStorage->processStorage();
			thunderDrive->processDrive();
			thunderLight->processLights();
            Wait(0.005);                // wait for a motor update time
		}
		thunderDrive->stopAction();
		thunderShooter->shootStop();
		thunderShooter->stopTurret();
	}

	void OperatorControl(void)
	{
		thunderShooter->requestZero();
		while (IsOperatorControl() && IsEnabled())
		{
			thunderControls->processBrokenSwitches();
			thunderControls->runControls();
			thunderControls->processDebug();
			thunderControls->processDashboard();
			thunderShooter->processShooter();
			thunderStorage->processStorage();
			thunderLight->processLights();
			Wait(0.005);				// wait for a motor update time
		}
	}
};

START_ROBOT_CLASS(ThunderBot);
